#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.commands`` -- Commands
"""

import sys
import platform

import edbob
from edbob import commands

import rattail


class Command(commands.Command):
    """
    The primary command for Rattail.
    """
    
    name = 'rattail'
    version = rattail.__version__
    description = "Retail Software Framework"
    long_description = """
Rattail is a retail software framework.

Copyright (c) 2010-2012 Lance Edgar <lance@edbob.org>

This program comes with ABSOLUTELY NO WARRANTY.  This is free software,
and you are welcome to redistribute it under certain conditions.
See the file COPYING.txt for more information.
"""


class DatabaseSyncCommand(commands.Subcommand):
    """
    Interacts with the database synchronization service; called as ``rattail
    dbsync``.
    """

    name = 'dbsync'
    description = "Manage the database synchronization service"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        if sys.platform == 'linux2':
            parser.add_argument('-p', '--pidfile',
                                help="Path to PID file", metavar='PATH')

    def run(self, args):
        from rattail.db.sync import linux as dbsync

        if args.subcommand == 'start':
            dbsync.start_daemon(args.pidfile)

        elif args.subcommand == 'stop':
            dbsync.stop_daemon(args.pidfile)


class FileMonitorCommand(commands.Subcommand):
    """
    Interacts with the file monitor service; called as ``rattail filemon``.
    This command expects a subcommand; one of the following:

    * ``rattail filemon start``
    * ``rattail filemon stop``

    On Windows platforms, the following additional subcommands are available:

    * ``rattail filemon install``
    * ``rattail filemon uninstall`` (or ``rattail filemon remove``)

    .. note::
       The Windows Vista family of operating systems requires you to launch
       ``cmd.exe`` as an Administrator in order to have sufficient rights to
       run the above commands.

    .. See :doc:`howto.use_filemon` for more information.
    """

    name = 'filemon'
    description = "Manage the file monitor service"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        if sys.platform == 'linux2':
            parser.add_argument('-p', '--pidfile',
                                help="Path to PID file", metavar='PATH')

        elif sys.platform == 'win32':
            install = subparsers.add_parser('install', help="Install service")
            install.set_defaults(subcommand='install')
            install.add_argument('-a', '--auto-start', action='store_true',
                                 help="Configure service to start automatically")
            remove = subparsers.add_parser('remove', help="Uninstall (remove) service")
            remove.set_defaults(subcommand='remove')
            uninstall = subparsers.add_parser('uninstall', help="Uninstall (remove) service")
            uninstall.set_defaults(subcommand='remove')

    def run(self, args):
        if sys.platform == 'linux2':
            from rattail.filemon import linux as filemon

            if args.subcommand == 'start':
                filemon.start_daemon(args.pidfile)

            elif args.subcommand == 'stop':
                filemon.stop_daemon(args.pidfile)

        elif sys.platform == 'win32':
            from edbob import win32
            from rattail.filemon import win32 as filemon
            from rattail.win32 import require_elevation

            require_elevation()

            # Execute typical service command.
            options = []
            if args.subcommand == 'install' and args.auto_start:
                options = ['--startup', 'auto']
            win32.execute_service_command(filemon, args.subcommand, *options)

            # If installing auto-start service on Windows 7, we should update
            # its startup type to be "Automatic (Delayed Start)".
            # TODO: Improve this check to include Vista?
            if args.subcommand == 'install' and args.auto_start:
                if platform.release() == '7':
                    name = filemon.RattailFileMonitor._svc_name_
                    win32.delayed_auto_start_service(name)

        else:
            sys.stderr.write("File monitor is not supported on platform: {0}\n".format(sys.platform))
            sys.exit(1)


class LoadHostDataCommand(commands.Subcommand):
    """
    Loads data from the Rattail host database, if one is configured.
    """

    name = 'load-host-data'
    description = "Load data from host database"

    def run(self, args):
        from edbob.console import Progress
        from rattail.db import load

        edbob.init_modules(['edbob.db'])

        if 'host' not in edbob.engines:
            print "Host engine URL not configured."
            return

        proc = load.LoadProcessor()
        proc.load_all_data(edbob.engines['host'], Progress)


class MakeConfigCommand(commands.Subcommand):
    """
    Creates a sample configuration file.
    """

    name = 'make-config'
    description = "Create a configuration file"

    def add_parser_args(self, parser):
        parser.add_argument('path', default='rattail.conf', metavar='PATH',
                            help="Path to the new file")
        parser.add_argument('-f', '--force', action='store_true',
                            help="Overwrite an existing file")


    def run(self, args):
        import os
        import os.path
        import shutil
        from rattail.files import resource_path

        dest = os.path.abspath(args.path)
        if os.path.exists(dest):
            if os.path.isdir(dest):
                sys.stderr.write("Path must include the filename; "
                                 "you gave a directory:\n  {0}\n".format(dest))
                sys.exit(1)
            if not args.force:
                sys.stderr.write("File already exists "
                                 "(use --force to overwrite):\n  "
                                 "{0}\n".format(dest))
                sys.exit(1)
            os.remove(dest)

        src = resource_path('rattail:data/rattail.conf.sample')
        shutil.copyfile(src, dest)


class MakeUserCommand(commands.Subcommand):
    """
    Creates a system user for Rattail.
    """

    name = 'make-user'
    description = "Create a system user account for Rattail"

    def run(self, args):
        from getpass import getpass
        from rattail.win32 import users

        if users.user_exists('rattail'):
            sys.stderr.write("User already exists: rattail\n")
            sys.exit(1)

        try:
            password = None
            while not password:
                password = getpass("Enter a password: ").strip()
        except KeyboardInterrupt:
            sys.stderr.write("Operation canceled by user.")
            sys.exit(2)

        users.create_rattail_user(password)
        sys.stdout.write("Created user: rattail\n")


class PalmCommand(commands.Subcommand):
    """
    Manages registration for the HotSync Manager conduit; called as::

       rattail palm
    """

    name = 'palm'
    description = "Manage the HotSync Manager conduit registration"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        register = subparsers.add_parser('register', help="Register Rattail conduit")
        register.set_defaults(subcommand='register')

        unregister = subparsers.add_parser('unregister', help="Unregister Rattail conduit")
        unregister.set_defaults(subcommand='unregister')

    def run(self, args):
        from rattail import palm
        from rattail.win32 import require_elevation
        from rattail.exceptions import PalmError

        require_elevation()

        if args.subcommand == 'register':
            try:
                palm.register_conduit()
            except PalmError, error:
                sys.stderr.write(str(error) + '\n')

        elif args.subcommand == 'unregister':
            try:
                palm.unregister_conduit()
            except PalmError, error:
                sys.stderr.write(str(error) + '\n')
                

class PurgeBatchesCommand(commands.Subcommand):
    """
    .. highlight:: sh

    Purges stale batches from the database; called as::

      rattail purge-batches
    """

    name = 'purge-batches'
    description = "Purge stale batches from the database"

    def add_parser_args(self, parser):
        parser.add_argument('-A', '--all', action='store_true',
                            help="Purge ALL batches regardless of purge date")

    def run(self, args):
        from rattail.db.batches.util import purge_batches

        edbob.init_modules(['edbob.db', 'rattail.db'])

        print "Purging batches from database:"
        print "    %s" % edbob.engine.url

        session = edbob.Session()
        purged = purge_batches(session, purge_everything=args.all)
        session.commit()
        session.close()

        print "\nPurged %d batches" % purged


def main(*args):
    """
    The primary entry point for the Rattail command system.
    """

    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)




import unittest

from pyramid import testing


class TestCase(unittest.TestCase):
    """
    Base class for all test suites.
    """

    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()

    def test_something(self):
        self.assertTrue(1)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db`` -- Database Stuff
"""

import logging
import warnings

from sqlalchemy.event import listen
from sqlalchemy.orm import object_mapper, RelationshipProperty

import edbob
from edbob.db import Session

import rattail


ignore_role_changes = None

log = logging.getLogger(__name__)


def before_flush(session, flush_context, instances):
    """
    Listens for session flush events.  This function is responsible for adding
    stub records to the 'changes' table, which will in turn be used by the
    database synchronizer.
    """

    def ensure_uuid(instance):
        if instance.uuid:
            return

        # If the 'uuid' column is actually a foreign key to another
        # table...well, then we can't just generate a new value for it.
        # Instead we must traverse the relationship and fetch the existing
        # foreign key value...

        mapper = object_mapper(instance)
        uuid = mapper.columns['uuid']
        if uuid.foreign_keys:

            for prop in mapper.iterate_properties:
                if (isinstance(prop, RelationshipProperty)
                    and len(prop.remote_side) == 1
                    and list(prop.remote_side)[0].key == 'uuid'):

                    foreign_instance = getattr(instance, prop.key)
                    ensure_uuid(foreign_instance)
                    instance.uuid = foreign_instance.uuid
                    break
            assert instance.uuid

        # ...but if there is no foreign key, just generate a new UUID.
        else:
            instance.uuid = edbob.get_uuid()

    def record_change(instance, deleted=False):

        # No need to record changes for Change. :)
        if isinstance(instance, rattail.Change):
            return

        # No need to record changes for batch data.
        if isinstance(instance, (rattail.Batch,
                                 rattail.BatchColumn,
                                 rattail.BatchRow)):
            return

        # Ignore instances which don't use UUID.
        if not hasattr(instance, 'uuid'):
            return

        # Ignore Role instances, if so configured.
        if ignore_role_changes:
            if isinstance(instance, (edbob.Role, edbob.UserRole)):
                return

        # Provide an UUID value, if necessary.
        ensure_uuid(instance)

        # Record the change.
        change = session.query(rattail.Change).get(
            (instance.__class__.__name__, instance.uuid))
        if not change:
            change = rattail.Change(
                class_name=instance.__class__.__name__,
                uuid=instance.uuid)
            session.add(change)
        change.deleted = deleted
        log.debug("before_flush: recorded change: %s" % repr(change))

    for instance in session.deleted:
        log.debug("before_flush: deleted instance: %s" % repr(instance))
        record_change(instance, deleted=True)

    for instance in session.new:
        log.debug("before_flush: new instance: %s" % repr(instance))
        record_change(instance)

    for instance in session.dirty:
        if session.is_modified(instance, passive=True):
            log.debug("before_flush: dirty instance: %s" % repr(instance))
            record_change(instance)


def record_changes(session):
    listen(session, 'before_flush', before_flush)


def init(config):
    """
    Initialize the Rattail database framework.
    """

    # Pretend all ``edbob`` models come from Rattail, until that is true.
    from edbob.db import Base
    names = []
    for name in edbob.__all__:
        obj = getattr(edbob, name)
        if isinstance(obj, type) and issubclass(obj, Base):
            names.append(name)
    edbob.graft(rattail, edbob, names)

    # Pretend all ``edbob`` enumerations come from Rattail.
    from edbob import enum
    edbob.graft(rattail, enum)

    from rattail.db.extension import model
    edbob.graft(rattail, model)

    global ignore_role_changes
    ignore_role_changes = config.getboolean(
        'rattail.db', 'changes.ignore_roles', default=True)

    if config.getboolean('rattail.db', 'changes.record'):
        record_changes(Session)

    elif config.getboolean('rattail.db', 'record_changes'):
        warnings.warn("Config setting 'record_changes' in section [rattail.db] "
                      "is deprecated; please use 'changes.record' instead.",
                      DeprecationWarning)
        record_changes(Session)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.batches.util`` -- Batch Utilities
"""

import re

from sqlalchemy import MetaData

import edbob
from edbob.time import local_time

import rattail
from rattail.db.extension.model import Batch


def purge_batches(session, effective_date=None, purge_everything=False):
    """
    Purge old batches from the database.

    :param session: Active database session.

    :param effective_date: Date against which comparisons should be made when
      determining if a batch is "old" (based on its ``purge_date`` attribute).
      The current date is assumed if none is specified.

    :param purge_everything: Boolean indicating whether the purge should be
      complete and indiscriminate, as opposed to honoring batch purge dates.
      (Use with caution!)

    :returns: Number of batches purged.
    :rtype: int
    """

    if effective_date is None:
        edbob.init_modules(['edbob.time'])
        effective_date = local_time().date()

    purged = 0

    q = session.query(Batch)
    if not purge_everything:
        q = q.filter(Batch.purge != None)
        q = q.filter(Batch.purge < effective_date)
    for batch in q:
        batch.drop_table()
        session.delete(batch)
        session.flush()
        purged += 1

    # This should theoretically not be necessary, if/when the batch processing
    # is cleaning up after itself properly.  For now though, it seems that
    # orphaned data tables are sometimes being left behind.

    batch_pattern = re.compile(r'^batch\.[0-9a-f]{32}$')

    current_batches = []
    for batch in session.query(Batch):
        current_batches.append('batch.%s' % batch.uuid)

    def orphaned_batches(name, metadata):
        if batch_pattern.match(name):
            if name not in current_batches:
                return True
        return False

    metadata = MetaData(session.bind)
    metadata.reflect(only=orphaned_batches)
    for table in reversed(metadata.sorted_tables):
        table.drop()

    return purged



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.diffs`` -- Data Diffs
"""

from rattail.db import Session
from rattail.db.sync import get_sync_engines

from sqlalchemy.orm import class_mapper


def instances_differ(local_instance, remote_instance, mapper):
    """
    Perform "shallow" comparison of data between two model instances.
    """

    for key in mapper.columns.keys():
        if getattr(local_instance, key) != getattr(remote_instance, key):
            return True

    return False


def find_diffs(local_session, remote_session, class_, progress=None):
    """
    Find differences between local and remote database, for a given model.
    """

    diffs = {}
    local_query = local_session.query(class_)
    remote_query = remote_session.query(class_)
    mapper = class_mapper(class_)

    count = local_query.count()
    if not count:
        return diffs

    prog = None
    if progress:
        prog = progress("Finding diffs for {0}".format(class_.__name__), count)

    local_uuids = []

    for i, local_instance in enumerate(local_query, 1):
        local_uuids.append(local_instance.uuid)
        remote_instance = remote_query.get(local_instance.uuid)

        if remote_instance is None:
            diffs.setdefault('missing_remote', [])
            diffs['missing_remote'].append(local_instance)

        elif instances_differ(local_instance, remote_instance, mapper):
            diffs.setdefault('different', [])
            diffs['different'].append((local_instance, remote_instance))

        if prog:
            prog.update(i)
    if prog:
        prog.destroy()

    count = remote_query.count()
    if not count:
        return diffs

    prog = None
    if progress:
        prog = progress("Finding diffs for {0}".format(class_.__name__), count)

    for i, remote_instance in enumerate(remote_query, 1):
        if remote_instance.uuid not in local_uuids:
            diffs.setdefault('missing_local', [])
            diffs['missing_local'].append(remote_instance)

        if prog:
            prog.update(i)
    if prog:
        prog.destroy()

    return diffs



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.batches.data`` -- Batch Data Providers
"""

import csv

from edbob.files import count_lines


__all__ = ['ProductQueryDataProvider', 'CSVDataProxy', 'CSVDataProvider']


class BatchDataProvider(object):

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError


class QueryDataProvider(BatchDataProvider):

    def __init__(self, query):
        self.query = query

    def __len__(self):
        return self.query.count()

    def __iter__(self):
        for data in self.query:
            yield data


class ProductDataProxy(object):

    def __init__(self, product):
        self.product = product

    def __getattr__(self, name):
        if name == 'F01':
            return self.product.upc
        if name == 'F02':
            return self.product.description
        if name == 'F22':
            return self.product.size
        if name == 'F155':
            if self.product.brand:
                return self.product.brand.name
            return ''
        raise AttributeError("Product has no attribute '%s'" % name)


class ProductQueryDataProvider(QueryDataProvider):

    def __iter__(self):
        for product in self.query:
            yield ProductDataProxy(product)


class CSVDataProxy(object):

    def __init__(self, row):
        self.row = row

    def __getattr__(self, name):
        if name in self.row:
            return self.row[name]
        raise AttributeError("CSV data row has no attribute '%s'" % name)


class CSVDataProvider(BatchDataProvider):

    proxy_class = CSVDataProxy

    def __init__(self, csv_path):
        self.csv_path = csv_path

    def __len__(self):
        return count_lines(self.csv_path)

    def __iter__(self):
        csv_file = open(self.csv_path, 'rb')
        reader = csv.DictReader(csv_file)
        for row in reader:
            yield self.proxy_class(row)
        csv_file.close()



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################


import os.path
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
execfile(os.path.join(here, 'rattail', '_version.py'))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()


requires = [
    #
    # Version numbers within comments below have specific meanings.
    # Basically the 'low' value is a "soft low," and 'high' a "soft high."
    # In other words:
    #
    # If either a 'low' or 'high' value exists, the primary point to be
    # made about the value is that it represents the most current (stable)
    # version available for the package (assuming typical public access
    # methods) whenever this project was started and/or documented.
    # Therefore:
    #
    # If a 'low' version is present, you should know that attempts to use
    # versions of the package significantly older than the 'low' version
    # may not yield happy results.  (A "hard" high limit may or may not be
    # indicated by a true version requirement.)
    #
    # Similarly, if a 'high' version is present, and especially if this
    # project has laid dormant for a while, you may need to refactor a bit
    # when attempting to support a more recent version of the package.  (A
    # "hard" low limit should be indicated by a true version requirement
    # when a 'high' version is present.)
    #
    # In any case, developers and other users are encouraged to play
    # outside the lines with regard to these soft limits.  If bugs are
    # encountered then they should be filed as such.
    #
    # package                           # low                   high

    'edbob[filemon]>=0.1a22',           #                       0.1a22
    ]


setup(
    name = "rattail",
    version = __version__,
    author = "Lance Edgar",
    author_email = "lance@edbob.org",
    url = "http://rattail.edbob.org/",
    license = "GNU Affero GPL v3",
    description = "Retail Software Framework",
    long_description = README + '\n\n' +  CHANGES,

    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Office/Business',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

    install_requires = requires,
    tests_require = requires + ['nose', 'coverage'],
    test_suite = 'nose.collector',

    namespace_packages = ['rattail'],
    packages = find_packages(),
    include_package_data = True,
    zip_safe = False,

    entry_points = """

[console_scripts]
rattail = rattail.commands:main

[gui_scripts]
rattailw = rattail.commands:main

[edbob.db.extensions]
rattail = rattail.db.extension:RattailExtension

[rattail.commands]
dbsync = rattail.commands:DatabaseSyncCommand
filemon = rattail.commands:FileMonitorCommand
load-host-data = rattail.commands:LoadHostDataCommand
make-config = rattail.commands:MakeConfigCommand
make-user = rattail.commands:MakeUserCommand
palm = rattail.commands:PalmCommand
purge-batches = rattail.commands:PurgeBatchesCommand

[rattail.batches.providers]
print_labels = rattail.batches.providers.labels:PrintLabels

[rattail.sil.column_providers]
rattail = rattail.sil.columns:provide_columns

""",
    )



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.sync`` -- Database Synchronization
"""

import sys
import time
import logging

if sys.platform == 'win32':
    import win32api

import sqlalchemy.exc
from sqlalchemy.orm import class_mapper

import edbob

import rattail
from rattail.db.model import CustomerGroupAssignment


log = logging.getLogger(__name__)


def get_sync_engines():
    edbob.init_modules(['edbob.db'])

    keys = edbob.config.get('rattail.db', 'syncs')
    if not keys:
        return None

    engines = {}
    for key in keys.split(','):
        key = key.strip()
        engines[key] = edbob.engines[key]
    log.debug("get_sync_engines: Found engine keys: %s" % ','.join(engines.keys()))
    return engines


def dependency_sort(x, y):
    map_x = class_mapper(getattr(edbob, x))
    map_y = class_mapper(getattr(edbob, y))

    dep_x = []
    table_y = map_y.tables[0].name
    for column in map_x.columns:
        for key in column.foreign_keys:
            if key.column.table.name == table_y:
                return 1
            dep_x.append(key)

    dep_y = []
    table_x = map_x.tables[0].name
    for column in map_y.columns:
        for key in column.foreign_keys:
            if key.column.table.name == table_x:
                return -1
            dep_y.append(key)

    if dep_x and not dep_y:
        return 1
    if dep_y and not dep_x:
        return -1
    return 0


class Synchronizer(object):
    """
    Default implementation of database synchronization logic.  Subclass this if
    you have special processing needs.
    """

    def synchronize_changes(self, engines):

        log.info("synchronize_changes: Using engine keys: %s" % ','.join(engines.keys()))

        while True:
            try:
                local_session = edbob.Session()
                local_changes = local_session.query(rattail.Change)

                if local_changes.count():
                    log.debug("synchronize_changes: found %d changes" % local_changes.count())

                    # First we must determine which types of instances are in need of
                    # syncing.  The order will matter because of class dependencies.
                    # However the dependency_sort() call doesn't *quite* take care of
                    # everything - notably the Product/ProductPrice situation.  Since
                    # those classes are mutually dependent, we start with a hackish
                    # lexical sort and hope for the best...
                    q = local_session.query(rattail.Change.class_name.distinct())
                    q = q.order_by('class_name')
                    class_names = [x[0] for x in q]
                    class_names.sort(cmp=dependency_sort)

                    remote_sessions = []
                    for remote_engine in engines.itervalues():
                        remote_sessions.append(
                            edbob.Session(bind=remote_engine))

                    for class_name in class_names:

                        for change in local_changes.filter_by(class_name=class_name):
                            log.debug("synchronize_changes: processing change: %s" % change)
                            cls = getattr(edbob, change.class_name)

                            if change.deleted:
                                for remote_session in remote_sessions:
                                    remote_instance = remote_session.query(cls).get(change.uuid)
                                    if remote_instance:
                                        self.delete_instance(remote_session, remote_instance)
                                        remote_session.flush()

                            else: # new/dirty
                                local_instance = local_session.query(cls).get(change.uuid)
                                if local_instance:
                                    for remote_session in remote_sessions:
                                        self.merge_instance(remote_session, local_instance)
                                        remote_session.flush()

                            local_session.delete(change)
                            local_session.flush()

                    for remote_session in remote_sessions:
                        remote_session.commit()
                        remote_session.close()
                    local_session.commit()

                local_session.close()
                self.sleep(3)

            except sqlalchemy.exc.OperationalError, error:
                if error.connection_invalidated:
                    # Presumably a database server restart; give it a moment
                    # and try again.
                    self.sleep(5)
                else:
                    raise

    def sleep(self, seconds):
        if sys.platform == 'win32':
            win32api.Sleep(seconds * 1000)
        else:
            time.sleep(seconds)

    def merge_instance(self, session, instance):
        """
        Merge ``instance`` into ``session``.

        This method checks for other "special" methods based on the class of
        ``instance``.  If such a method is found, it is invoked; otherwise a
        simple merge is done.
        """

        cls = instance.__class__.__name__
        if hasattr(self, 'merge_%s' % cls):
            return getattr(self, 'merge_%s' % cls)(session, instance)

        return session.merge(instance)

    def merge_Product(self, session, source_product):
        """
        This method is somewhat of a hack, in order to properly handle
        :class:`rattail.Product` instances and the interdependent nature of the
        related :class:`rattail.ProductPrice` instances.
        """

        target_product = session.merge(source_product)

        # I'm not 100% sure I understand this correctly, but here's my
        # thinking: First we clear the price relationships in case they've
        # actually gone away; then we re-establish any which are currently
        # valid.

        # Setting the price relationship attributes to ``None`` isn't enough to
        # force the ORM to notice a change, since the UUID field is ultimately
        # what it's watching.  So we must explicitly use that field here.
        target_product.regular_price_uuid = None
        target_product.current_price_uuid = None

        # If the source instance has currently valid price relationships, then
        # we re-establish them.  We must merge the source price instance in
        # order to be certain it will exist in the target session, and avoid
        # foreign key errors.  However we *still* must also set the UUID fields
        # because again, the ORM is watching those...  This was noticed to be
        # the source of some bugs where successive database syncs were
        # effectively "toggling" the price relationship.  Setting the UUID
        # field explicitly seems to solve it.
        if source_product.regular_price_uuid:
            target_product.regular_price = session.merge(source_product.regular_price)
            target_product.regular_price_uuid = target_product.regular_price.uuid
        if source_product.current_price_uuid:
            target_product.current_price = session.merge(source_product.current_price)
            target_product.current_price_uuid = target_product.current_price.uuid

        return target_product

    def delete_instance(self, session, instance):
        """
        Delete ``instance`` using ``session``.

        This method checks for other "special" methods based on the class of
        ``instance``.  If such a method is found, it is invoked before the
        instance is officially deleted from ``session``.
        """

        cls = instance.__class__.__name__
        if hasattr(self, 'delete_%s' % cls):
            getattr(self, 'delete_%s' % cls)(session, instance)

        session.delete(instance)

    def delete_Department(self, session, department):

        # Remove association from Subdepartment records.
        q = session.query(rattail.Subdepartment)
        q = q.filter(rattail.Subdepartment.department == department)
        for subdepartment in q:
            subdepartment.department = None

        # Remove association from Product records.
        q = session.query(rattail.Product)
        q = q.filter(rattail.Product.department == department)
        for product in q:
            product.department = None

    def delete_Subdepartment(self, session, subdepartment):

        # Remove association from Product records.
        q = session.query(rattail.Product)
        q = q.filter(rattail.Product.subdepartment == subdepartment)
        for product in q:
            product.subdepartment = None

    def delete_Vendor(self, session, vendor):

        # Remove associated ProductCost records.
        q = session.query(rattail.ProductCost)
        q = q.filter(rattail.ProductCost.vendor == vendor)
        for cost in q:
            session.delete(cost)

    def delete_CustomerGroup(self, session, group):
        # First remove customer associations.
        q = session.query(CustomerGroupAssignment)\
            .filter(CustomerGroupAssignment.group == group)
        for assignment in q:
            session.delete(assignment)


def synchronize_changes(engines):
    """
    This function will instantiate a ``Synchronizer`` class according to
    configuration.  (The default class is :class:`Synchronizer`.)  This
    instance is then responsible for implementing the sync logic.

    .. highlight:: ini

    If you need to override the default synchronizer class, put something like
    the following in your config file::

       [rattail.db]
       sync.synchronizer_class = myapp.sync:MySynchronizer
    """

    cls = edbob.config.get('rattail.db', 'sync.synchronizer_class')
    if cls:
        cls = edbob.load_spec(cls)
    else:
        cls = Synchronizer
    sync = cls()

    sync.synchronize_changes(engines)
